# -*- coding: utf-8 -*-
"""
Created on Wed Aug 30 17:15:43 2017

@author: mwojc
"""
from fempy.geometry.gmsh import Gmsh
from fempy.geometry.tools import tmpfname
import sys
sys.path.append('/usr/lib64')
import FreeCAD, Part, Draft


def Circles(R=1., r=0.85, shift=(0, 0, 0), hole=True):
    d = FreeCAD.newDocument()
    cout = Draft.makeCircle(R)
    p = FreeCAD.Placement()
    p.Base = shift
    cin = Draft.makeCircle(r, p)
    geometry = cout.Shape.cut(cin.Shape)
    if not hole:
        geometry = Part.Compound([geometry, cin.Shape])
    fname = tmpfname() + '.brep'
    geometry.exportBrep(fname)
    g = Gmsh(fname)
    g.options.Mesh.CharacteristicLengthFactor = 0.25
    g.options.Mesh.ElementOrder = 2
    return g


def Balls(R=1., r=0.85, shift=(0, 0, 0), hole=True):
    d = FreeCAD.newDocument()
    b1 = Part.makeSphere(R)
    b2 = Part.makeSphere(r, FreeCAD.Vector(*shift))
    geometry = b1.cut(b2)
    if not hole:
        geometry = Part.Compound([geometry, b2])
    fname = tmpfname() + '.brep'
    geometry.exportBrep(fname)
    g = Gmsh(fname)
    g.options.Mesh.CharacteristicLengthFactor = 0.25
    g.options.Mesh.ElementOrder = 2
    return g


Circles(shift=(1, 0, 0)).preview()
Balls(shift=(0, 0, 1)).preview()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#try:
from setuptools import setup, findall
#except ImportError:
    #from distutils.core import setup


if __name__ == "__main__":
    setup(name              = 'fempy',
          version           = '0.3.1',
          description       = 'Finite element method in python',
          long_description  = open('README.md', 'r').read(),
          long_description_content_type='text/markdown',
          keywords          = ['FEM, fem, finite element method'],
          author            = 'Marek Wojciechowski',
          author_email      = 'mwojc@p.lodz.pl',
          url               = 'http://fempy.org',
          license           = 'LGPL-3',
          platforms         = 'Posix, Windows',
          package_dir       = {'fempy': 'fempy',
                               'fempy.geometry': 'fempy/geometry',
                               'fempy.domain': 'fempy/domain',
                               'fempy.fields': 'fempy/fields',
                               'fempy.models': 'fempy/models',
                               'fempy.solvers': 'fempy/solvers',
                               'fempy.tests': 'tests',
                               'fempy.tests.geometry': 'tests/geometry',
                               'fempy.tests.domain': 'tests/domain',
                               'fempy.tests.fields': 'tests/fields',
                               'fempy.tests.models': 'tests/models',
                               'fempy.tests.solvers': 'tests/solvers',
                               'fempy.examples': 'examples'},
          packages          = ['fempy',
                               'fempy.geometry',
                               'fempy.domain',
                               'fempy.fields',
                               'fempy.models',
                               'fempy.solvers',
                               'fempy.tests',
                               'fempy.tests.geometry',
                               'fempy.tests.domain',
                               'fempy.tests.fields',
                               'fempy.tests.models',
                               'fempy.tests.solvers',
                               'fempy.examples'],
          package_data      = {'fempy.tests': ['data/*',
                                               'data/gmsh/*',
                                               'data/elastic/*',
                                               'data/heat/*']},
          #packages = find_packages(),
          #package_data      = {'tests': ['tests/data/*']},
          install_requires  = ['numpy', 'scipy'],
          classifiers       = ['Development Status :: 4 - Beta',
                               'Intended Audience :: Education',
                               'Intended Audience :: Science/Research',
                               'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
                               'Operating System :: POSIX :: Linux',
                               'Operating System :: Microsoft :: Windows',
                               'Operating System :: MacOS :: MacOS X',
                               'Programming Language :: C',
                               'Programming Language :: C++',
                               'Programming Language :: Python :: Implementation :: CPython',
                               'Topic :: Scientific/Engineering'],
          )



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 22 22:47:46 2014

@author: mwojc
"""

import numpy as np
from fempy.domain.domain import Domain
from fempy.domain.domain import Volume, Surface, Curve, Curve2D, Surface3D, Curve3D
from fempy.testing import testdata
from common import Simple2D, Simple3D
import pytest


class TestDomain(object):
    def call_attributes(self, e):
        e.faces
        e.boundary
        e.ndime
        e.nnode
        e.nedim
        e.nelem
        e.nenod
        e.ndeg
        e.coors
        e.conec
        e.ecoors
        e.gcoors
        e.gjac
        e.gdet
        e.gvol
        e.evol
        e.vol
        e.gijac
        e.egn
        e.egnk


class TestVolume(TestDomain):
    def test_tetr4(self):
        coors = np.loadtxt(testdata + 'cylinder4n.coors')
        conec = np.loadtxt(testdata + 'cylinder4n.conec', dtype=int)
        c = Volume((coors, conec))  # , 'tetr4')
        self.call_attributes(c)
        d = Domain(coors, conec)
        assert isinstance(d, Volume)
        self.call_attributes(d)

    def test_tetr10(self):
        coors = np.loadtxt(testdata + 'cylinder10n.coors')
        conec = np.loadtxt(testdata + 'cylinder10n.conec', dtype=int)
        c = Volume((coors, conec), 'tetr10')
        self.call_attributes(c)
        d = Domain(coors, conec)
        assert isinstance(d, Volume)
        self.call_attributes(d)


class TestSurface(TestDomain):
    def test_tria3(self):
        coors = np.loadtxt(testdata + 'circle3n.coors')
        conec = np.loadtxt(testdata + 'circle3n.conec', dtype=int)
        c = Surface((coors, conec), 'tria3')
        self.call_attributes(c)
        d = Domain(coors, conec)
        assert isinstance(d, Surface)
        self.call_attributes(d)

    def test_tria6(self):
        coors = np.loadtxt(testdata + 'circle6n.coors')
        conec = np.loadtxt(testdata + 'circle6n.conec', dtype=int)
        c = Surface((coors, conec), 'tria6')
        self.call_attributes(c)
        d = Domain(coors, conec)
        assert isinstance(d, Surface)
        self.call_attributes(d)


class TestCurve(TestDomain):
    def test_line2(self):
        coors = np.loadtxt(testdata + 'line2n.coors')
        coors.shape += (1,)
        conec = np.loadtxt(testdata + 'line2n.conec', dtype=int)
        c = Curve((coors, conec), 'line2')
        self.call_attributes(c)
        d = Domain(coors, conec)
        assert isinstance(d, Curve)
        self.call_attributes(d)

    def test_line3(self):
        coors = np.loadtxt(testdata + 'line3n.coors')
        coors.shape += (1,)
        conec = np.loadtxt(testdata + 'line3n.conec', dtype=int)
        c = Curve((coors, conec), 'line3')
        self.call_attributes(c)
        d = Domain(coors, conec)
        assert isinstance(d, Curve)
        self.call_attributes(d)


class TestCurve2D(TestDomain):
    def test_line2(self):
        coors = np.loadtxt(testdata + 'circum2n.coors')
        conec = np.loadtxt(testdata + 'circum2n.conec', dtype=int)
        c = Curve2D((coors, conec), 'line2')
        self.call_attributes(c)
        # More attributes
        c.tangent
        c.normal
        c.gjac
        d = Domain(coors, conec)
        assert isinstance(d, Curve2D)
        self.call_attributes(d)

    def test_line3(self):
        coors = np.loadtxt(testdata + 'circum3n.coors')
        conec = np.loadtxt(testdata + 'circum3n.conec', dtype=int)
        c = Curve2D((coors, conec), 'line3')
        self.call_attributes(c)
        c.tangent
        c.normal
        c.gjac
        d = Domain(coors, conec)
        assert isinstance(d, Curve2D)
        self.call_attributes(d)


class TestSurface3D(TestDomain):
    def test_tria3(self):
        coors = np.loadtxt(testdata + 'paraboloid3n.coors')
        conec = np.loadtxt(testdata + 'paraboloid3n.conec', dtype=int)
        c = Surface3D((coors, conec), 'tria3')
        self.call_attributes(c)
        c.tangent
        c.normal
        c.gjac
        d = Domain(coors, conec)
        assert isinstance(d, Surface3D)
        self.call_attributes(d)

    def test_tria6(self):
        coors = np.loadtxt(testdata + 'paraboloid6n.coors')
        conec = np.loadtxt(testdata + 'paraboloid6n.conec', dtype=int)
        c = Surface3D((coors, conec), 'tria6')
        self.call_attributes(c)
        c.tangent
        c.normal
        c.gjac
        d = Domain(coors, conec)
        assert isinstance(d, Surface3D)
        self.call_attributes(d)


class TestCurve3D(TestDomain):
    def test_line2(self):
        coors = np.loadtxt(testdata + 'spiral2n.coors')
        conec = np.loadtxt(testdata + 'spiral2n.conec', dtype=int)
        c = Curve3D((coors, conec), 'line2')
        self.call_attributes(c)
        c.tangent
        c.normal
        c.orient
        c.gjac
        d = Domain(coors, conec)
        assert isinstance(d, Curve3D)
        self.call_attributes(d)

    def test_line3(self):
        coors = np.loadtxt(testdata + 'spiral3n.coors')
        conec = np.loadtxt(testdata + 'spiral3n.conec', dtype=int)
        c = Curve3D((coors, conec), 'line3')
        self.call_attributes(c)
        c.tangent
        c.normal
        c.orient
        c.gjac
        d = Domain(coors, conec)
        assert isinstance(d, Curve3D)
        self.call_attributes(d)


class TestSimpleDomain2D(Simple2D):
    def test_boundary(self):
        c1 = self.conec1D[:-1]  # without last element
        c2 = self.domain2D.boundary.eidx.conec_g
        b1 = np.lexsort(c1.T)
        b2 = np.lexsort(c2.T)
        assert np.allclose(c1[b1], c2[b2])

    def test_indexing(self):
        allc = np.allclose
        nnz = np.nonzero
        d = self.domain2D
        # b = self.boundary
        assert allc(nnz(d.eidx['q2']), [2, 3])
        assert allc(nnz(d.eidx.nidx['q2']), [1, 2, 4, 5])
        assert allc(d['q2'].eidx.lst, [2, 3])
        assert allc(d['q2'].eidx.nidx.lst, [1, 2, 4, 5])
        assert d.eidx.nidx.g is d['q2'].eidx.nidx.g
        assert d.eidx.g is d['q2'].eidx.g

    def test_subdomain(self):
        allc = np.allclose
        nnz = np.nonzero
        d = self.domain2D
        d.faces.egroups['left'] = [1, 3, 4]
        q4 = d['q4']
        assert d.faces.eidx.g is q4.faces.eidx.g
        assert allc(nnz(q4.eidx.msk), [6, 7])
        assert allc(nnz(q4.eidx.nidx.msk), [4, 5, 7, 8])
        assert allc(q4.eidx.lst[q4.egroups['boundary']], [6, 7])
        assert allc(q4.eidx.lst[q4.egroups['BOUNDARY']], [7])
        assert allc(q4.eidx.lst[q4.egroups['CUT']], [6])
        assert allc(q4.faces.coors['left'], d.coors['left*q4'])
        assert allc(q4.boundary.coors['left'], d.coors['left*q4'])
        assert allc(q4.coors['q2'], d.coors['q2*q4'])
        assert allc(q4.boundary.vol, 4.)
        assert allc(q4.vol, 1.)
        assert allc(q4.boundary['q2*q4'].conec.shape[0], 1)
        assert allc(q4.faces['q2*q4'].conec.shape[0], 1)

    def test_ngroups_numbers(self):
        from fempy.fields.nfields import nScalar
        d = self.domain2D
        u = nScalar(d, np.arange(d.nnode))
        assert len(u) == 9

        b = d.boundary
        bu = nScalar(b, np.arange(b.nnode))
        assert len(bu) == 8

        bu[:] = 1
        uu = u + bu
        assert len(uu) == 9
        assert uu[0] == 1
        assert uu[4] == 4
        assert uu[5] == 6

        u[:] = 3
        buu = bu + u
        assert len(buu) == 8
        assert np.allclose(buu, 4)

        u += bu
        assert len(u) == 9
        assert u[0] == 4
        assert u[4] == 3
        assert u[5] == 4

        bu += u
        assert len(bu) == 8
        assert np.allclose(bu, 5)

    def test_ngroups_nums_indices(self):
        allc = np.allclose
        d = self.domain2D
        assert allc(d.ngroups.nums('l'), [0, 3, 6])
        assert allc(d.ngroups.nums('l', reverse=True), [6, 3, 0])
        assert allc(d.ngroups.nums('l', 'y', reverse=False), [0, 3, 6])
        assert allc(d.ngroups.nums('cross'), [2, 4, 6])
        assert allc(d.ngroups.nums('cross', 'x'), [6, 4, 2])
        assert allc(d.ngroups.nums('cross', 'y'), [2, 4, 6])
        assert allc(d.ngroups.nums('cross', 'xy'), [2, 4, 6])
        assert allc(d.ngroups.nums('cross', 'yx'), [6, 4, 2])

    def test_cached_properties(self):
        d = self.domain2D
        f = d.faces
        b = d.boundary
        d.coors
        d.gvol
        d.boundary.normal
        assert d.coors is d.__dict__['coors']
        assert d.gijac is d.__dict__['gijac']
        assert d.boundary.normal is d.boundary.__dict__['normal']
        d.invalidate_caches()
        with pytest.raises(KeyError):
            d.__dict__['coors']
        with pytest.raises(KeyError):
            d.__dict__['gijac']
        with pytest.raises(KeyError):
            d.boundary.__dict__['normal']
        assert f is d.faces
        assert b is d.boundary

    def test_update(self):
        d = self.domain2D
        coors = d.coors
        disp = np.random.rand(*coors.shape)*0.01
        d.update(disp)
        assert np.allclose(d.coors, coors+disp)
        d.update(-disp)  # recover original coors

    def test_axisymmetry(self):
        allc = np.allclose
        nnz = np.nonzero
        d = self.domain2D
        b = d.boundary

        d.axis = 'auto'
        assert allc(d.axis, 0.)
        assert allc(d.radius, d.gcoors[..., 0])
        assert allc(d.radiusn, d.coors[..., 0])
        assert allc(d.gvol, 2*np.pi*d.radius*(0.5*1*1))
        assert allc(nnz(d.ngroups['axis']), [0, 3, 6])
        assert allc(b.axis, 0.)
        assert allc(b.radius, b.gcoors[..., 0])
        assert allc(b.radiusn, b.coors[..., 0])
        assert allc(b.gvol, 2*np.pi*b.radius*(1))
        assert allc(nnz(b.eidx.nidx.g['axis']), [0, 3, 6])

        d.axis = -1.
        assert allc(d.axis, -1.)
        assert allc(d.radius, d.gcoors[..., 0]+1.)
        assert allc(d.radiusn, d.coors[..., 0]+1.)
        assert allc(d.gvol, 2*np.pi*d.radius*(0.5*1*1))
        assert allc(d.ngroups['axis'], False)
        assert allc(b.axis, -1)
        assert allc(b.radius, b.gcoors[..., 0]+1)
        assert allc(b.radiusn, b.coors[..., 0]+1)
        assert allc(b.gvol, 2*np.pi*b.radius*(1))
        assert allc(b.eidx.nidx.g['axis'], False)

        d.axis = None
        assert d.axis is None
        assert d.radius is None
        assert d.radiusn is None
        assert allc(d.gvol, (0.5*1*1))
        with pytest.raises(KeyError):
            d.ngroups['axis']
        with pytest.raises(AssertionError):
            d.axis = 1.

    def test_thickness(self):
        d = self.domain2D
        gvol = d.gvol
        d.thickness[:] = 0.1
        assert np.allclose(gvol, 10*d.gvol)
        # assert np.allclose(d.boundary.thickness, 0.1)

        d = self.domain1D
        gvol = d.gvol
        d.thickness[:] = 0.1
        assert np.allclose(gvol, 10*d.gvol)
        d.width[:] = 0.1
        assert np.allclose(gvol, 100*d.gvol)
        # assert np.allclose(d.boundary.thickness, 0.1)


class TestSimple3D(Simple3D):
    def test_boundary(self):
        idx = []
        for k in ['left', 'right', 'bottom', 'top', 'front', 'back']:
            idx += self.egroups2D[k]
        c1 = self.conec2D[idx]
        c2 = self.domain3D.boundary.eidx.conec_g
        c1.sort(1)
        c2.sort(1)
        b1 = np.lexsort(c1.T)
        b2 = np.lexsort(c2.T)
        assert np.allclose(c1[b1], c2[b2])


if __name__ == '__main__':
    # import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 11:17:33 2016

@author: mwojc
"""
from fempy.compat import pickle
from fempy.models.heat import Heat
from fempy.solvers.tools import model_test
from fempy.testing import testdata
import numpy as np


class Test_Heat(object):
    def test_heat_2D_p2(self):
        d = pickle.load(open(testdata + '/heat/2D_p2.dom', 'rb'))
        m = Heat(d)

        m.kij[:] = [[1., 0], [0, 1.]]

        m.alpha['left + bottom'] = 8.
        m.epsilon['left + bottom'] = 0.85
        m.Text['left + bottom'] = 293.15

        m.alpha['right + top'] = 25.
        m.epsilon['right + top'] = 0.85
        m.Text['right + top'] = 1273.15

        model_test(m, x0=np.loadtxt(testdata + '/heat/2D_p2.sol'), strict=True)

    def test_heat_2D_s1(self):
        d = pickle.load(open(testdata + '/heat/2D_s1.dom', 'rb'))
        m = Heat(d)

        m.kij[:] = [[1., 0], [0, 1.]]

        m.alpha['left'] = 8.
        m.epsilon['left'] = 0.85
        m.Text['left'] = 293.15

        m.alpha['right'] = 25.
        m.epsilon['right'] = 0.85
        m.Text['right'] = 1273.15

        model_test(m, x0=np.loadtxt(testdata + '/heat/2D_s1.sol'), strict=True)

    def test_heat_3D_p2(self):
        d = pickle.load(open(testdata + '/heat/3D_p2.dom', 'rb'))
        m = Heat(d)

        m.kij[:] = [[1., 0, 0], [0, 1., 0], [0, 0, 1.]]

        m.alpha['left+bottom'] = 8.
        m.epsilon['left+bottom'] = 0.85
        m.Text['left+bottom'] = 293.15

        m.alpha['right+top'] = 25.
        m.epsilon['right+top'] = 0.85
        m.Text['right+top'] = 1273.15

        model_test(m, x0=np.loadtxt(testdata + '/heat/3D_p2.sol'), strict=True)

    def test_heat_3D_s1(self):
        d = pickle.load(open(testdata + '/heat/3D_s1.dom', 'rb'))
        m = Heat(d)

        m.kij[:] = [[1., 0, 0], [0, 1., 0], [0, 0, 1.]]

        m.alpha['left'] = 8.
        m.epsilon['left'] = 0.85
        m.Text['left'] = 293.15

        m.alpha['right'] = 25.
        m.epsilon['right'] = 0.85
        m.Text['right'] = 1273.15

        model_test(m, x0=np.loadtxt(testdata + '/heat/3D_s1.sol'), strict=True)

    def test_heat_3D(self):
        d = pickle.load(open(testdata + '/heat/3D.dom', 'rb'))
        m = Heat(d)

        m.kij[:] = [[1., 0, 0], [0, 1., 0], [0, 0, 1.]]

        m.alpha['left+bottom+front'] = 8.
        m.epsilon['left+bottom+front'] = 0.85
        m.Text['left+bottom+front'] = 293.15

        m.alpha['right+top+back'] = 25.
        m.epsilon['right+top+back'] = 0.85
        m.Text['right+top+back'] = 1273.15

        model_test(m, x0=np.loadtxt(testdata + '/heat/3D.sol'), strict=True)

    def test_heat_1D(self):
        d = pickle.load(open(testdata + '/heat/1D.dom', 'rb'))
        m = Heat(d)

        m.kij[:] = [[1.]]

        m.alpha['left'] = 8.
        m.epsilon['left'] = 0.85
        m.Text['left'] = 293.15

        m.alpha['right'] = 25.
        m.epsilon['right'] = 0.85
        m.Text['right'] = 1273.15

        model_test(m, x0=np.loadtxt(testdata + '/heat/1D.sol'), strict=True)


if __name__ == '__main__':
    import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 27 21:30:46 2014

@author: mwojc
"""

from fempy.geometry import Rectangle
from fempy.fields.gfields import gScalar, gTensor
from fempy.fields.nfields import nScalar
from fempy.fields.operators import dot, grad, nabla, qform, lform, dirichlet
from fempy.solvers.linear import dsolve
from fempy.geometry.gmsh_io import preview_fields


d = Rectangle(elsize_factor=0.5).gmsh.domain(order=2)

# Assemble stiffness matrix
p = nScalar(d)
kij = gTensor(d, [[1., 0], [0, 1.]])
K = qform(kij, nabla(p))

# Assemble flux
q = gScalar(d.boundary)
q['left + bottom'] = 1.
q = -lform(q)

# Dirichlet BC's - prescribed pressures
r = dirichlet(p)
r['top + right'] = True
p['top + right'] = 1.

# Solve
p = dsolve(K, q, r, p)

# Retrieve velocities
v = -dot(kij, grad(p))

# Preview fields
preview_fields([p, v], ['p [kPa]', 'v [m/s]'])

# FempyView is depreciated
#### View
#from fempy.view import FempyView
#view = FempyView(d, mesh_visible=False)
#view.point_scalars(P, name='p')
#view.point_vectors(onnodes(V), name='v',
#                   scale_factor=0.05,
#                   color_mode='no_color')



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 19 11:15:18 2014

@author: mwojc
"""
import numpy as np
from fempy.fields.operators import onnodes, onelems, ongauss, onboundary
from fempy.fields.operators import grad, div, integrate, lform, qform
from fempy.fields.operators import operator, nabla
from fempy.fields.operators import small_strain, small_rotation, divergence
from fempy.fields.algebra import transpose
from fempy.fields.nfields import nField, nScalar, nVector, nList
from fempy.fields.nfields import nTensor, nMatrix, nTensor3, nTensor4
from fempy.fields.gfields import gField, gScalar, gVector, gList
from fempy.fields.gfields import gTensor, gMatrix, gTensor3, gTensor4
from fempy.testing import testdata
from common import Simple2D


class Test_Operators_nField:
    coors = np.loadtxt(testdata + 'paraboloid6n.coors')
    conec = np.loadtxt(testdata + 'paraboloid6n.conec', dtype=int)
    from fempy.domain.domain import Surface3D
    domain = Surface3D((coors, conec), 'tria6')

    def get_fields(self):
        d = self.domain
        s = nScalar(d)
        v = nVector(d)
        l = nList(d, [1, 2, 3, 4, 5])
        t = nTensor(d)
        m = nMatrix(d, (2, 3))
        t3 = nTensor3(d)
        t4 = nTensor4(d, 1.)
        gf = m.view(nField)
        return s, v, l, t, m, t3, t4, gf

    def get_dims(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime
        ng = d.etype.ngauss
        ne = d.nelem
        return nn, nd, ne, ng

    def test_onboundary(self):
        d = Simple2D().domain2D
        u = nVector(d, 1.)
        ub = nVector(d.boundary, 1.)
        np.allclose(onboundary(u), ub)
        u = nVector(d, np.random.rand(9, 2))
        np.allclose(onboundary(u), u['boundary'])
        np.allclose(u['BOUNDARY'], u['boundary'])
        d = self.domain
        u = nVector(d)
        u[:] = np.random.rand(*u.shape)
        np.allclose(onboundary(u), u['boundary'])
        np.allclose(u['BOUNDARY'], u['boundary'])

    def test_onnodes(self):
        nn, nd, ne, ng = self.get_dims()
        fields = self.get_fields()

        for f in fields:
            assert np.allclose(f, onnodes(f))

    def test_ongauss(self):
        nn, nd, ne, ng = self.get_dims()
        fields = self.get_fields()
        s, v, l, t, m, t3, t4, gf = [ongauss(f) for f in fields]

        assert s.shape == (ne, ng)
        assert v.shape == (ne, ng, nd)
        assert l.shape == (ne, ng, 5)
        assert t.shape == (ne, ng, nd, nd)
        assert m.shape == (ne, ng, 2, 3)
        assert t3.shape == (ne, ng, nd, nd, nd)
        assert t4.shape == (ne, ng, nd, nd, nd, nd)
        assert type(s) == gScalar
        assert type(v) == gVector
        assert type(l) == gList
        assert type(t) == gTensor
        assert type(m) == gMatrix
        assert type(t3) == gTensor3
        assert type(t4) == gTensor4

        assert np.allclose(t4, 1.)

    def test_grad(self):
        nn, nd, ne, ng = self.get_dims()
        fields = self.get_fields()
        s, v, l, t, m, t3, t4, gf = [grad(f) for f in fields]

        assert s.shape == (ne, ng, nd)
        assert v.shape == (ne, ng, nd, nd)
        assert l.shape == (ne, ng, 5, nd)
        assert t.shape == (ne, ng, nd, nd, nd)
        assert m.shape == (ne, ng, 2, 3, nd)
        assert t3.shape == (ne, ng, nd, nd, nd, nd)
        assert t4.shape == (ne, ng, nd, nd, nd, nd, nd)
        assert type(s) == gVector
        assert type(v) == gTensor
        assert type(l) == gMatrix
        assert type(t) == gTensor3
        assert type(m) == gField
        assert type(t3) == gTensor4
        assert type(t4) == gField

        assert np.allclose(t4, 0.)

    def test_div(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4, gf = self.get_fields()
        v, t, t3, t4 = [div(f) for f in (v, t, t3, t4)]

        assert v.shape == (ne, ng)
        assert t.shape == (ne, ng, nd)
        assert t3.shape == (ne, ng, nd, nd)
        assert t4.shape == (ne, ng, nd, nd, nd)
        assert type(v) == gScalar
        assert type(t) == gVector
        assert type(t3) == gTensor
        assert type(t4) == gTensor3

        import pytest
        for f in s, l, m:
            with pytest.raises(AssertionError):
                div(f)

        assert np.allclose(t4, 0.)

    def test_integrate(self):
        nn, nd, ne, ng = self.get_dims()
        fields = self.get_fields()
        s, v, l, t, m, t3, t4, gf = [integrate(grad(f)) for f in fields]

        assert s.shape == (nn,)
        assert v.shape == (nn, nd)
        assert l.shape == (nn, 5)
        assert t.shape == (nn, nd, nd)
        assert m.shape == (nn, 2, 3)
        assert t3.shape == (nn, nd, nd, nd)
        assert t4.shape == (nn, nd, nd, nd, nd)
        assert type(s) == nScalar
        assert type(v) == nVector
        assert type(l) == nList
        assert type(t) == nTensor
        assert type(m) == nMatrix
        assert type(t3) == nTensor3
        assert type(t4) == nTensor4

        assert np.allclose(t4, 0.)
        assert np.allclose(integrate(t4).sum(), 0.)

    def test_operator(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4, gf = self.get_fields()

        for f in s, v, l, t, m, t3, t4:
            f[:] = np.random.rand(*f.shape)
            i = 'ijklmopq'[:f.sufix]
            j = 'rstuvwxy'[:f.sufix]
            feg = np.einsum('egn%s%s,en%s->eg%s' % (i, j, j, i),
                            operator(f), onelems(f))
            assert np.allclose(feg, ongauss(f))

    def test_nabla(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4, gf = self.get_fields()

        for f in s, v, l, t, m, t3, t4:
            i = 'ijklmopq'[:f.sufix]
            j = 'rstuvwxy'[:f.sufix]
            fegk = np.einsum('egn%s%sd,en%s->eg%sd'
                             % (i, j, j, i), nabla(f), onelems(f))
            assert np.allclose(fegk, grad(f))

    def test_small_strain(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4, gf = self.get_fields()

        v[:] = np.random.rand(*v.shape)
        g = grad(v)
        ss1 = 0.5*(g + transpose(g))
        ss2 = small_strain(v).gfield
        assert np.allclose(ss1, ss2)

        import pytest
        for f in s, l, t, m, t3, t4:
            with pytest.raises(AssertionError):
                small_strain(f)

    def test_small_rotation(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4, gf = self.get_fields()

        v[:] = np.random.rand(*v.shape)
        g = grad(v)
        ss1 = 0.5*(g - transpose(g))
        ss2 = small_rotation(v).gfield
        assert np.allclose(ss1, ss2)

        import pytest
        for f in s, l, t, m, t3, t4:
            with pytest.raises(AssertionError):
                small_rotation(f)

    def test_divergence(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4, gf = self.get_fields()

        for f in v, t, t3, t4:
            f[:] = np.random.rand(*f.shape)
            d1 = div(f)
            d2 = divergence(f).gfield
            assert np.allclose(d1, d2)

        import pytest
        for f in s, l, m:
            with pytest.raises(AssertionError):
                divergence(f)

    def test_lform(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4, gf = self.get_fields()

        s[:] = 1.
        assert np.allclose(lform(s).sum(), s.domain.vol)
        v[:] = 1.
        assert np.allclose(lform(v).sum(), v.domain.vol*nd)
        t[:] = 1.
        assert np.allclose(lform(t).sum(), t.domain.vol*nd**2)
        l[:] = 1.
        assert np.allclose(lform(l).sum(), l.domain.vol*5)
        m[:] = 1.
        assert np.allclose(lform(m).sum(), m.domain.vol*6)
        # assert type(lform(m)) == nMatrix
        t3[:] = 1.
        assert np.allclose(lform(t3).sum(), t3.domain.vol*nd**3)
        t4[:] = 1.
        assert np.allclose(lform(t4).sum(), t4.domain.vol*nd**4)

        v[:] = np.random.rand(*v.shape)
        V = lform(v)
        r = V.sum(0)
        for i in range(nd):
            s[:] = v[..., i]
            assert np.allclose(lform(s).sum(), r[i])

    def test_qform(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4, gf = self.get_fields()

        u = nabla(s)
        A = t
        K = qform(A, u)
        assert K.shape == (nn,)*2

        u = nabla(v)
        A = t4
        K = qform(A, u)
        assert K.shape == (nn*nd,)*2

        u = nabla(t)
        A = gField(t.domain, np.random.rand(*(nd,)*6))  # sixth order
        K = qform(A, u)
        assert K.shape == (nn*nd*nd,)*2


class Test_Operators_gField:
    coors = np.loadtxt(testdata + 'paraboloid6n.coors')
    conec = np.loadtxt(testdata + 'paraboloid6n.conec', dtype=int)
    from fempy.domain.domain import Surface3D
    domain = Surface3D((coors, conec), 'tria6')

    def get_fields(self):
        d = self.domain
        s = gScalar(d)
        v = gVector(d)
        l = gList(d, [1, 2, 3, 4, 5])
        t = gTensor(d)
        m = gMatrix(d, (2, 3))
        t3 = gTensor3(d)
        t4 = gTensor4(d, 1.)
        return s, v, l, t, m, t3, t4

    def get_dims(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime
        ng = d.etype.ngauss
        ne = d.nelem
        return nn, nd, ne, ng

    def test_ongauss(self):
        nn, nd, ne, ng = self.get_dims()
        fields = self.get_fields()

        for f in fields:
            assert np.allclose(f, ongauss(f))

    def test_onnodes(self):
        nn, nd, ne, ng = self.get_dims()
        fields = self.get_fields()
        s, v, l, t, m, t3, t4 = [onnodes(f) for f in fields]

        assert s.shape == (nn,)
        assert v.shape == (nn, nd)
        assert l.shape == (nn, 5)
        assert t.shape == (nn, nd, nd)
        assert m.shape == (nn, 2, 3)
        assert t3.shape == (nn, nd, nd, nd)
        assert t4.shape == (nn, nd, nd, nd, nd)
        assert type(s) == nScalar
        assert type(v) == nVector
        assert type(l) == nList
        assert type(t) == nTensor
        assert type(m) == nMatrix
        assert type(t3) == nTensor3
        assert type(t4) == nTensor4

        assert np.allclose(t4, 1.)

    def test_integrate(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4 = self.get_fields()
        v, t, t4 = [integrate(f) for f in (v, t, t4)]

        assert v.shape == (nn,)
        assert t.shape == (nn, nd)
        assert t4.shape == (nn, nd, nd, nd)
        assert type(v) == nScalar
        assert type(t) == nVector
        assert type(t4) == nTensor3

        assert np.allclose(t4.sum(), 0)

    def test_lform_default(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4 = self.get_fields()

        s[:] = 1.
        assert np.allclose(lform(s).sum(), s.domain.vol)
        v[:] = 1.
        assert np.allclose(lform(v).sum(), v.domain.vol*nd)
        t[:] = 1.
        assert np.allclose(lform(t).sum(), t.domain.vol*nd**2)
        l[:] = 1.
        assert np.allclose(lform(l).sum(), l.domain.vol*5)
        m[:] = 1.
        assert np.allclose(lform(m).sum(), m.domain.vol*6)
        # assert type(lform(m)) == nMatrix
        t3[:] = 1.
        assert np.allclose(lform(t3).sum(), t3.domain.vol*nd**3)
        t4[:] = 1.
        assert np.allclose(lform(t4).sum(), t4.domain.vol*nd**4)

        v[:] = np.random.rand(*v.shape)
        V = lform(v)
        r = V.sum(0)
        for i in range(nd):
            s[:] = v[..., i]
            np.allclose(lform(s).sum(), r[i])

    def test_qform_default(self):
        nn, nd, ne, ng = self.get_dims()
        s, v, l, t, m, t3, t4 = self.get_fields()

        A = t
        K = qform(A)
        assert K.shape == (nn,)*2

        A = t4
        K = qform(A)
        assert K.shape == (nn*nd,)*2

        A = gField(t.domain, np.random.rand(*(nd,)*6))  # sixth order
        K = qform(A)
        assert K.shape == (nn*nd*nd,)*2

    def test_assemble(self):
        d = Simple2D().domain2D
        dt = d['q3+q4']
        u = nVector(d)
        cijkl = gTensor4(d, 1.)
        K1 = qform(cijkl, nabla(u))
        ut = nVector(dt)
        cijklt = gTensor4(dt, 1.)
        K2 = qform(cijklt, nabla(ut))
        ub = nVector(d.boundary)
        cijklb = gTensor4(d.boundary, 1.)
        K3 = qform(cijklb, nabla(ub))
        K = K1 + K2 + K3
        assert K.shape == (18, 18)

        L1 = lform(u)
        assert L1.shape == (9, 2)

    def test_assemble_subdomain(self):
        d = Simple2D().domain2D
        dt = d['q3+q4']
        from fempy.domain.domain import Domain
        # CHANGE SUPERDOMAIN AGINST WHICH THINGS ARE ASSEMBLED
        dt2 = Domain(dt.coors, dt.conec, etype=dt.etype, egroups=dt.egroups,
                     ngroups=dt.ngroups)
        ut = nVector(dt2)
        cijklt = gTensor4(dt2, 1.)
        K2 = qform(cijklt, nabla(ut))
        ub = nVector(dt2.boundary)
        cijklb = gTensor4(dt2.boundary, 1.)
        K3 = qform(cijklb, nabla(ub))
        K = K2 + K3
        assert K.shape == (12, 12)

    def test_assemble_orphaned_point(self):
        s = Simple2D()
        from fempy.domain.index import NIndex, EIndex
        from fempy.domain.domain import Surface
        nidx = NIndex(s.coors)
        eidx = EIndex(s.conec2D, nidx=nidx)
        sd = Surface(eidx)
        # sd = Simple2D().domain2D
        # sd.coors = np.concatenate((sd.domain.coors, [[3, 3]]))  # add point
        u = nVector(sd)
        cijkl = gTensor4(sd, 1.)
        K1 = qform(cijkl, nabla(u))
        assert K1.shape == (20, 20)

        L1 = lform(u)
        assert L1.shape == (10, 2)


if __name__ == '__main__':
    import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 27 21:30:46 2014

@author: mwojc
"""

from fempy.geometry import Rectangle
from fempy.models.elastic import ev2cijkl
from fempy.fields.nfields import nVector
from fempy.fields.gfields import gScalar, gVector, gTensor, gTensor4
from fempy.fields.operators import dot, grad, nabla
from fempy.fields.operators import qform, lform, dirichlet
from fempy.fields.algebra import transpose, tensordot
from fempy.solvers.linear import dsolve  # direct solver
from fempy.geometry.gmsh_io import preview_fields

d = Rectangle(elsize=0.05).gmsh.domain(2)

# Assemble stiffness matrix
u = nVector(d)
nu = nabla(u)
bu = 0.5*(nu + transpose(nu))  # small strain operator
cijkl = gTensor4(d, ev2cijkl(20000., 0.3, ndime=2))  # material tensor
K = qform(cijkl, bu)  # the same trial and test

# Assemble body forces - weight
F = nVector(d)
rho = gScalar(d, 2.)
gravity = gVector(d, [0., -9.81])
weight = dot(rho, gravity)
F += lform(weight)

# Assemble boundary pressures (given vector)
q = gVector(d.boundary)
q['left'] = [10, 0.]
F += lform(q)  # operator(onboundary(u)))

# Assemble boundary pressures (given scalar normal pressure)
p = gScalar(d.boundary)
p['right'] = -10.
n = d.boundary.normal
f = dot(p, n)
F += lform(f)  # operator(onboundary(u)))

# Assemble thermal load
alpha = 1e-4 * gTensor(d, [[1., 0], [0, 1]])
dT = 3.
eps_dT = alpha * dT
sigma_dT = (1 + 0.3)*tensordot(cijkl, eps_dT)  # 0.3 is Poisson coefficient
F += lform(sigma_dT, bu)

# Dirichlet BC's
r = dirichlet(u)
u['top', 1] = -0.0005
r['top', 1] = True
r['bottom'] = True

# Solve (linear)
u = dsolve(K, F, r, u)

# Strains and stresses
du = grad(u)
eps = 0.5*(du + transpose(du))
sigma = tensordot(cijkl, eps - eps_dT)

# View
preview_fields([u, sigma], ['u [m]', 'stress [kPa]'])

# FempyView is currently depreciated
#from fempy.view import FempyView
#view = FempyView(d, u*500, mesh_visible=True)
#view.cell_scalars(sigma[..., 0, 1, 1], name='Syy', legend=True)

#view.point_scalars(onnodes(sigma)[:, 1, 1], name='Syy')
#view.point_scalars(onboundary(onnodes(sigma))[:, 1, 1], name='Syy')
#view.add_lines(d.boundary.coors, d.boundary.conec)
#view.cell_scalars(sigma[:, 0, 1, 1], name='Syy', legend=True)



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 19 11:25:01 2014

@author: mwojc
"""
import numpy as np
from fempy.fields.gfields import gField, gScalar, gVector, gList
from fempy.fields.gfields import gTensor, gMatrix, gTensor3, gTensor4
from fempy.fields.algebra import dot, tensordot
from fempy.testing import testdata
from common import Simple3D
import pytest


class Test_gField(Simple3D):
    @property
    def domain(self):
        return self.domain3D

    def test_getitem(self):
        d = self.domain
        f = gTensor(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        for k, v in d.egroups.items():
            assert (v == f.groups[k]).all()
        d.egroups['test2'] = slice(1, 6, 2)
        assert (f['test2'] == f[1:6:2]).all()

        with pytest.raises(AttributeError):
            f.groups = {}
        eg = d.egroups
        d.egroups = {}
        assert f.groups == {}
        d.egroups = eg

    def test_setitem(self):
        d = self.domain
        f = gTensor(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        f['bottome'] = 5.
        assert np.allclose(f['bottome', 0, 0], [[5., 5., 5.]]*4)

    def test_scalar_field_creation(self):
        d = self.domain
        ne = d.nelem
        ng = d.etype.ngauss
        f = gScalar(d)
        assert type(f) == gScalar
        assert f.shape == (ne, ng)
        assert np.allclose(f, 0.)
        assert f.domain is d
        f = gScalar(d, 1.)
        assert type(f) == gScalar
        assert f.shape == (ne, ng)
        assert np.allclose(f, 1.)
        f = gScalar(d, f)
        assert type(f) == gScalar
        with pytest.raises(AssertionError):
            f = gScalar(d, [2.])
        with pytest.raises(AssertionError):
            f = gScalar(d, [[2.], [1]])
        f = gScalar(d, 'test')
        assert f.dtype in ('S4', '<U4')

    def test_vector_field_creation(self):
        d = self.domain
        ne = d.nelem
        ng = d.etype.ngauss
        nd = d.ndime

        f = gVector(d)
        assert type(f) == gVector
        assert f.shape == (ne, ng, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = gVector(d, 1.)
        assert f.shape == (ne, ng, nd)
        assert np.allclose(f, 1.)
        assert f.domain is d

        f = gVector(d, [1., 1., 0.])
        assert f.shape == (ne, ng, nd)
        assert np.allclose(f[..., -1], 0.)
        assert f.domain is d

        f = gVector(d, f)
        assert f.shape == (ne, ng, nd)
        assert np.allclose(f[..., -1], 0.)
        assert f.domain is d

        with pytest.raises(AssertionError):
            f = gVector(d, [2., 2.])
        with pytest.raises(AssertionError):
            f = gVector(d, [[2.], [1]])
        f = gVector(d, 'test')
        assert f.dtype in ('S4', '<U4')

    def test_tensor_field_creation(self):
        d = self.domain
        ne = d.nelem
        ng = d.etype.ngauss
        nd = d.ndime

        f = gTensor(d)
        assert type(f) == gTensor
        assert f.shape == (ne, ng, nd, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = gTensor(d, 1.)
        assert f.shape == (ne, ng, nd, nd)
        assert np.allclose(f, 1.)
        assert f.domain is d

        f = gTensor(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        assert f.shape == (ne, ng, nd, nd)
        assert np.allclose(f[..., -1, 0], 2.)
        assert f.domain is d

        f = gTensor(d, f)
        assert f.shape == (ne, ng, nd, nd)
        assert np.allclose(f[..., 0, 0], 1.)
        assert f.domain is d

        with pytest.raises(AssertionError):
            f = gTensor(d, [[2., 2.], [1., 1.]])
        with pytest.raises(AssertionError):
            f = gTensor(d, [[2., 1]])
        f = gTensor(d, 3)
        assert f.dtype == int

    def test_tensor3_field_creation(self):
        d = self.domain
        ne = d.nelem
        ng = d.etype.ngauss
        nd = d.ndime

        f = gTensor3(d)
        assert type(f) == gTensor3
        assert f.shape == (ne, ng, nd, nd, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = gTensor3(d, 1.)
        assert f.shape == (ne, ng, nd, nd, nd)
        assert np.allclose(f, 1.)
        assert f.domain is d

        r = np.random.rand(3, 3, 3)
        f = gTensor3(d, r)
        assert f.shape == (ne, ng, nd, nd, nd)
        assert np.allclose(f[..., 1, 1, 1], r[1, 1, 1])
        assert f.domain is d

        f = gTensor3(d, f)
        assert f.shape == (ne, ng, nd, nd, nd)
        assert np.allclose(f[..., 0, 0, 0], r[0, 0, 0])
        assert f.domain is d

        with pytest.raises(AssertionError):
            f = gTensor3(d, [[2., 2.], [1., 1.]])
        with pytest.raises(AssertionError):
            f = gTensor3(d, [[2., 1]])
        f = gTensor3(d, 3)
        assert f.dtype == int

    def test_tensor4_field_creation(self):
        d = self.domain
        ne = d.nelem
        ng = d.etype.ngauss
        nd = d.ndime

        f = gTensor4(d)
        assert type(f) == gTensor4
        assert f.shape == (ne, ng, nd, nd, nd, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = gTensor4(d, 1.)
        assert f.shape == (ne, ng, nd, nd, nd, nd)
        assert np.allclose(f, 1.)
        assert f.domain is d

        r = np.random.rand(3, 3, 3, 3)
        f = gTensor4(d, r)
        assert f.shape == (ne, ng, nd, nd, nd, nd)
        assert np.allclose(f[..., 1, 1, 1, 1], r[1, 1, 1, 1])
        assert f.domain is d

        f = gTensor4(d, f)
        assert f.shape == (ne, ng, nd, nd, nd, nd)
        assert np.allclose(f[..., 0, 0, 0, 0], r[0, 0, 0, 0])
        assert f.domain is d

        with pytest.raises(AssertionError):
            f = gTensor4(d, [[2., 2.], [1., 1.]])
        with pytest.raises(AssertionError):
            f = gTensor4(d, [[2., 1]])
        f = gTensor4(d, 3)
        assert f.dtype == int

    def test_list_creation(self):
        d = self.domain
        ne = d.nelem
        ng = d.etype.ngauss
        nd = d.ndime

        f = gList(d, [1., 1., 0.])
        assert type(f) == gVector
        assert f.shape == (ne, ng, nd)
        assert np.allclose(f[..., -1], 0.)
        assert f.domain is d

        f = gList(d, gVector(d, [1., 1., 0.]))
        assert type(f) == gVector
        assert f.shape == (ne, ng, nd)
        assert np.allclose(f[..., -1], 0.)
        assert f.domain is d

        f = gList(d, 5)
        assert type(f) == gList
        assert f.shape == (ne, ng, 5)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = gList(d, [3, 2])
        assert type(f) == gList
        assert f.shape == (ne, ng, 2)
        assert np.allclose(f[..., 0], 3)
        assert f.domain is d
        assert f.dtype == int

        with pytest.raises(AssertionError):
            f = gList(d, [[2., 2.], [1., 1.]])
        with pytest.raises(AssertionError):
            f = gList(d, [[2., 1]])
        with pytest.raises(AssertionError):
            f = gList(d, 1.)

    def test_matrix_creation(self):
        d = self.domain
        ne = d.nelem
        ng = d.etype.ngauss
        nd = d.ndime

        f = gMatrix(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        assert type(f) == gTensor
        assert f.shape == (ne, ng, nd, nd)
        assert np.allclose(f[..., -1, 0], 2.)
        assert f.domain is d

        f = gMatrix(d, gTensor(d))
        assert type(f) == gTensor
        assert f.shape == (ne, ng, nd, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = gMatrix(d, (5, 2))
        assert type(f) == gMatrix
        assert f.shape == (ne, ng, 5, 2)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = gMatrix(d, [[3, 2], [2, 1]])
        assert type(f) == gMatrix
        assert f.shape == (ne, ng, 2, 2)
        assert np.allclose(f[..., 0, 0], 3)
        assert f.domain is d
        assert f.dtype == int

        with pytest.raises(AssertionError):
            f = gMatrix(d, [2., 1])
        with pytest.raises(AssertionError):
            f = gMatrix(d, [[[2., 1]]])
        with pytest.raises(AssertionError):
            f = gMatrix(d, 1.)

    def test_gfield_creation(self):
        d = self.domain
        ne = d.nelem
        ng = d.etype.ngauss
        nd = d.ndime

        with pytest.raises(TypeError):
            f = gField(d)

        f = gField(d, 2.)
        assert type(f) == gScalar

        f = gField(d, [1, 2, 3])
        assert type(f) == gVector
        assert f.dtype == int

        f = gField(d, [1, 2, 3, 4.])
        assert type(f) == gList

        f = gField(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        assert type(f) == gTensor

        f = gField(d, [[1., 1., 0.], [1., 1., 0.]])
        assert type(f) == gMatrix

        r = np.random.rand(3, 3, 3)
        f = gField(d, r)
        assert type(f) == gTensor3

        r = np.random.rand(3, 3, 3, 3)
        f = gField(d, r)
        assert type(f) == gTensor4

        r = np.random.rand(3, 2, 3, 4)
        f = gField(d, r)
        assert type(f) == gField

        r = np.random.rand(ne, ng, nd, nd)
        f = gField(d, r)
        assert type(f) == gTensor

    def test_nfield_creation(self):
        coors = np.loadtxt(testdata + 'paraboloid6n.coors')
        conec = np.loadtxt(testdata + 'paraboloid6n.conec', dtype=int)
        from fempy.domain.domain import Surface3D
        d = Surface3D((coors, conec), 'tria6')
        ne = d.nelem
        ng = d.etype.ngauss
        nd = d.ndime

        from fempy.fields.nfields import nVector
        v = nVector(d)
        f = gField(d, v)

        assert type(f) == gVector
        assert f.shape == (ne, ng, nd)

    def test_tensordot_dot(self):
        d = self.domain
        t4 = gTensor4(d, 1.)
        t2 = gTensor(d, 3.)
        v = gVector(d, 4.)

        assert type(tensordot(t4, t2)) == gTensor
        assert type(tensordot(t2, t2)) == gScalar
        assert type(dot(t2, v)) == gVector
        assert type(dot(v, v)) == gScalar
        assert type(dot(t2, t2)) == gTensor


if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 22 22:47:46 2014

@author: mwojc
"""

import numpy as np
from fempy.domain.index import GlobalNIndex, NIndex, GlobalEIndex, EIndex
from common import Simple2D
import pytest


class TestSimple2DIndex(Simple2D):
    """
      6 __7__ 8
       |\ |\ |
      3|_\4_\|5
       |\ |\ |
       |_\|_\|__
      0   1  2  9

    """
    def test_global_nidx_full(self, gi=None):
        nnz = np.nonzero
        allc = np.allclose
        if gi is None:
            gi = GlobalNIndex(self.coors)
            gi.update(self.ngroups)
        assert allc(nnz(gi[:]), np.arange(10))
        assert allc(nnz(gi['cline']), [3, 4, 5])
        assert allc(nnz(gi['cline*cpoint']), [4])
        assert allc(nnz(gi['~(cline*cpoint)']), [0, 1, 2, 3, 5, 6, 7, 8, 9])
        assert allc(nnz(gi['cpoint+opoint']), [4, 9])
        assert allc(nnz(gi['cline^cpoint']), [3, 5])
        assert allc(nnz(gi['x>1.01']), [2, 5, 8, 9])
        assert allc(nnz(gi['(x>1.01) * (y>1.01)']), [8])
        with pytest.raises(KeyError):
            gi['aaa']
        with pytest.raises(KeyError):
            gi['x']
        with pytest.raises(KeyError):
            gi['x*[1,2]']

    def test_global_nidx_conec2D(self, gi=None):
        nnz = np.nonzero
        allc = np.allclose
        if gi is None:
            gi = GlobalNIndex(self.coors, self.conec2D)
            gi.update(self.ngroups)
        assert len(gi[:]) == 10
        assert allc(nnz(gi[:]), np.arange(9))
        assert allc(gi.pos, [0, 1, 2, 3, 4, 5, 6, 7, 8, -1])
        assert allc(nnz(gi['cline']), [3, 4, 5])
        assert allc(nnz(gi['cline*cpoint']), [4])
        assert allc(nnz(gi['~(cline*cpoint)']), [0, 1, 2, 3, 5, 6, 7, 8])
        assert allc(nnz(gi['cpoint+opoint']), [4])
        assert allc(nnz(gi['cline^cpoint']), [3, 5])
        assert allc(nnz(gi['x>1.01']), [2, 5, 8])
        assert allc(nnz(gi['(x>1.01) * (y>1.01)']), [8])

    def test_global_nidx_conec1D(self, gi=None):
        nnz = np.nonzero
        allc = np.allclose
        if gi is None:
            gi = GlobalNIndex(self.coors, self.conec1D)
            gi.update(self.ngroups)
        assert allc(nnz(gi[:]), [0, 1, 2, 3, 5, 6, 7, 8, 9])
        assert allc(gi.pos, [0, 1, 2, 3, -1, 4, 5, 6, 7, 8])
        assert allc(nnz(gi['cline']), [3, 5])
        assert allc(nnz(gi['cline*cpoint']), [])
        assert allc(nnz(gi['~(cline*cpoint)']), [0, 1, 2, 3, 5, 6, 7, 8, 9])
        assert allc(nnz(gi['cpoint+opoint']), [9])
        assert allc(nnz(gi['cline^cpoint']), [3, 5])
        assert allc(nnz(gi['x>1.01']), [2, 5, 8, 9])
        assert allc(nnz(gi['(x>1.01) * (y>1.01)']), [8])

    def test_global_nidx_conec0D(self, gi=None):
        nnz = np.nonzero
        allc = np.allclose
        if gi is None:
            gi = GlobalNIndex(self.coors, self.conec0D)
            gi.update(self.ngroups)
        assert allc(nnz(gi[:]), [0, 2, 6, 8, 9])
        assert allc(gi.pos, [0, -1, 1, -1, -1, -1, 2, -1, 3, 4])
        assert allc(nnz(gi['cline']), [])
        assert allc(nnz(gi['cline*cpoint']), [])
        assert allc(nnz(gi['~(cline*cpoint)']), [0, 2, 6, 8, 9])
        assert allc(nnz(gi['cpoint+opoint']), [9])
        assert allc(nnz(gi['cline^cpoint']), [])
        assert allc(nnz(gi['x>1.01']), [2, 8, 9])
        assert allc(nnz(gi['(x>1.01) * (y>1.01)']), [8])

    def test_nidx_create_from_coors(self):
        allc = np.allclose

        ni = NIndex(self.coors)
        ni.g.update(self.ngroups)
        assert allc(ni.lst, ni.g.lst)
        assert set(ni.keys()) == set(ni.g.keys())
        self.test_global_nidx_full(ni.g)

        ni = NIndex(self.coors, self.conec1D)
        ni.g.update(self.ngroups)
        assert allc(ni.lst, ni.g.lst)
        assert set(ni.keys()) == set(ni.g.keys())
        self.test_global_nidx_conec1D(ni.g)

    def test_nidx_create_from_global(self):
        allc = np.allclose
        gi = GlobalNIndex(self.coors)
        gi.update(self.ngroups)
        ni = NIndex(gi)
        assert gi is ni.g
        assert allc(ni.lst, ni.g.lst)
        assert set(ni.keys()) == set(ni.g.keys())
        self.test_global_nidx_full(ni.g)

        ni = NIndex(gi, self.conec1D)
        assert gi is ni.g
        assert allc(ni.lst, [0, 1, 2, 3, 5, 6, 7, 8, 9])
        assert allc(ni.pos, [0, 1, 2, 3, -1, 4, 5, 6, 7, 8])
        assert set(ni.keys()) == set(ni.g.keys())
        self.test_global_nidx_full(ni.g)

    def test_nidx_create_from_local(self):
        gi = GlobalNIndex(self.coors)
        gi.update(self.ngroups)
        ni = NIndex(gi)
        with pytest.raises(ValueError):
            NIndex(ni, [0, 1, 2])

    def test_nidx_full(self, ni=None):
        if ni is None:
            ni = NIndex(self.coors)
            ni.g.update(self.ngroups)
        self.test_global_nidx_full(ni)

    def test_nidx_conec2D(self, ni=None):
        nnz = np.nonzero
        allc = np.allclose
        if ni is None:
            ni = NIndex(self.coors, self.conec2D)
            ni.g.update(self.ngroups)
        assert len(ni[:]) == 9
        assert allc(nnz(ni[:]), np.arange(9))
        assert allc(ni.pos, [0, 1, 2, 3, 4, 5, 6, 7, 8, -1])
        assert allc(nnz(ni['cline']), [3, 4, 5])
        assert allc(nnz(ni['cline*cpoint']), [4])
        assert allc(nnz(ni['~(cline*cpoint)']), [0, 1, 2, 3, 5, 6, 7, 8])
        assert allc(nnz(ni['cpoint+opoint']), [4])
        assert allc(nnz(ni['cline^cpoint']), [3, 5])
        assert allc(nnz(ni['x>1.01']), [2, 5, 8])
        assert allc(nnz(ni['(x>1.01) * (y>1.01)']), [8])

    def test_nidx_conec1Da(self, ni=None):
        nnz = np.nonzero
        allc = np.allclose
        if ni is None:
            ni = NIndex(self.coors, self.conec1D)
            ni.g.update(self.ngroups)
        assert allc(nnz(ni[:]), [0, 1, 2, 3, 4, 5, 6, 7, 8])
        assert allc(ni.pos, [0, 1, 2, 3, -1, 4, 5, 6, 7, 8])
        assert allc(ni.g.pos, [0, 1, 2, 3, -1, 4, 5, 6, 7, 8])
        assert allc(nnz(ni['cline']), [3, 4])
        assert allc(nnz(ni['~cline']), [0, 1, 2, 5, 6, 7, 8])
        assert allc(nnz(ni['cline*cpoint']), [])
        assert allc(nnz(ni['~(cline*cpoint)']), [0, 1, 2, 3, 4, 5, 6, 7, 8])
        assert allc(nnz(ni['cpoint+opoint']), [8])
        assert allc(nnz(ni['cline^cpoint']), [3, 4])
        assert allc(nnz(ni['x>1.01']), [2, 4, 7, 8])
        assert allc(nnz(ni['(x>1.01) * (y>1.01)']), [7])

    def test_nidx_conec1Db(self, ni=None):
        nnz = np.nonzero
        allc = np.allclose
        if ni is None:
            gi = GlobalNIndex(self.coors)
            gi.update(self.ngroups)
            ni = NIndex(gi, self.conec1D)
        assert allc(nnz(ni[:]), [0, 1, 2, 3, 4, 5, 6, 7, 8])
        assert allc(ni.pos, [0, 1, 2, 3, -1, 4, 5, 6, 7, 8])
        assert allc(ni.g.pos, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        assert allc(nnz(ni['cline']), [3, 4])
        assert allc(nnz(ni['~cline']), [0, 1, 2, 5, 6, 7, 8])
        assert allc(nnz(ni['cline*cpoint']), [])
        assert allc(nnz(ni['~(cline*cpoint)']), [0, 1, 2, 3, 4, 5, 6, 7, 8])
        assert allc(nnz(ni['cpoint+opoint']), [8])
        assert allc(nnz(ni['cline^cpoint']), [3, 4])
        assert allc(nnz(ni['x>1.01']), [2, 4, 7, 8])
        assert allc(nnz(ni['(x>1.01) * (y>1.01)']), [7])

    def test_nidx_conec0D(self, ni=None):
        nnz = np.nonzero
        allc = np.allclose
        if ni is None:
            gi = GlobalNIndex(self.coors)
            gi.update(self.ngroups)
            ni = NIndex(gi, self.conec0D)
        assert allc(nnz(ni[:]), [0, 1, 2, 3, 4])
        assert allc(ni.pos, [0, -1, 1, -1, -1, -1, 2, -1, 3, 4])
        assert allc(nnz(ni['cline']), [])
        assert allc(nnz(ni['cline*cpoint']), [])
        assert allc(nnz(ni['~(cline*cpoint)']), [0, 1, 2, 3, 4])
        assert allc(nnz(ni['cpoint+opoint']), [4])
        assert allc(nnz(ni['cline^cpoint']), [])
        assert allc(nnz(ni['x>1.01']), [1, 3, 4])
        assert allc(nnz(ni['(x>1.01) * (y>1.01)']), [3])

    def test_nidx_subindex(self):
        nnz = np.nonzero
        allc = np.allclose
        gi = GlobalNIndex(self.coors)
        gi.update(self.ngroups)
        ni = NIndex(gi, self.conec1D)
        subni = ni.subindex()
        allc(ni.lst, subni.lst)
        subni = ni.subindex('cline')
        allc(subni.lst, [3, 5])
        assert allc(nnz(subni[:]), [0, 1])
        subni = ni.subindex(gkey=self.conec0D)
        allc(subni.lst, [0, 2, 6, 8, 9])

        gi = GlobalNIndex(self.coors, self.conec2D)
        gi.update(self.ngroups)
        ni = NIndex(gi, self.conec1D)
        assert allc(ni.lst, [0, 1, 2, 3, 5, 6, 7, 8])
        subni = ni.subindex(gkey=self.conec0D)
        assert allc(subni.lst, [0, 2, 6, 8])

    def test_global_eidx_create(self):
        allc = np.allclose
        # create from coors as nindex
        gei = GlobalEIndex(self.conec2D, self.coors)
        gei.nidx.g.update(self.ngroups)
        self.test_global_nidx_conec2D(gei.nidx.g)
        self.test_nidx_conec2D(gei.nidx)
        # create from global nindex
        gni = GlobalNIndex(self.coors)
        gei = GlobalEIndex(self.conec2D, gni)
        gei.nidx.g.update(self.ngroups)
        self.test_global_nidx_full(gei.nidx.g)
        self.test_nidx_conec2D(gei.nidx)
        # create from local nindex
        ni = NIndex(self.coors, self.conec1D)
        gei = GlobalEIndex(self.conec2D, ni)
        gei.nidx.g.update(self.ngroups)
        self.test_global_nidx_conec1D(gei.nidx.g)
        assert allc(gei.nidx.lst, [0, 1, 3, 5, 7, 8])
        assert allc(gei.lst, [0, 7])  # do we really need this?

    def test_global_eidx_conec2D(self, gi=None):
        nnz = np.nonzero
        allc = np.allclose
        if gi is None:
            gi = GlobalEIndex(self.conec2D, self.coors)
            gi.nidx.g.update(self.ngroups)
            self.test_global_nidx_conec2D(gi.nidx.g)
            self.test_nidx_conec2D(gi.nidx)
            gi.update(self.egroups2D)
        assert len(gi[:]) == 8
        assert allc(nnz(gi[:]), np.arange(8))
        assert allc(gi.pos, [0, 1, 2, 3, 4, 5, 6, 7])
        assert allc(nnz(gi['q2']), [2, 3])
        assert allc(nnz(gi.nidx['q2']), [1, 2, 4, 5])
        assert allc(nnz(gi['q1+q2']), [0, 1, 2, 3])
        assert allc(nnz(gi['q3+q4']), [4, 5, 6, 7])
        assert allc(nnz(gi['~(q1+q2)']), [4, 5, 6, 7])
        assert allc(nnz(gi['q1*cline']), [])     # do we really need this?
        gi.nidx.g['conec1D'] = self.conec1D
        assert allc(nnz(gi['conec1D']), [0, 7])  # do we really need this?
        assert allc(nnz(gi['conec1D*q4']), [7])  # do we really need this?
        assert allc(nnz(gi['x>0.99']), [2, 3, 6, 7])
        assert allc(nnz(gi['(x>0.99) * (y>0.99)']), [6, 7])
        assert allc(nnz(gi['(x>0.99) * (y>0.99) * conec1D']), [7])

    def test_global_eidx_conec1D(self, gi=None):
        nnz = np.nonzero
        allc = np.allclose
        if gi is None:
            gi = GlobalEIndex(self.conec1D, self.coors)
            gi.nidx.g.update(self.ngroups)
            self.test_global_nidx_conec1D(gi.nidx.g)
            self.test_nidx_conec1Da(gi.nidx)
            gi.update(self.egroups1D)
            assert allc(nnz(gi.nidx.g['r*cline']), [5])
            assert allc(nnz(gi.nidx['r*cline']), [4])
        assert len(gi[:]) == 9
        assert allc(nnz(gi[:]), np.arange(9))
        assert allc(gi.pos, [0, 1, 2, 3, 4, 5, 6, 7, 8])
        with pytest.raises(KeyError):
            assert allc(nnz(gi['q2']), [2, 3])
        assert allc(nnz(gi['l']), [6, 7])
        assert allc(nnz(gi['t+l']), [4, 5, 6, 7])
        assert allc(nnz(gi['b+r']), [0, 1, 2, 3])
        assert allc(nnz(gi['~(b+r)']), [4, 5, 6, 7, 8])
        assert allc(nnz(gi['l*cline']), [])     # do we really need this?
        gi.nidx.g['conec0D'] = self.conec0D
        assert allc(nnz(gi['conec0D']), [8])  # do we really need this?
        assert allc(nnz(gi['conec0D*l']), [])  # do we really need this?
        assert allc(nnz(gi['x>0.99']), [1, 2, 3, 4, 8])
        assert allc(nnz(gi['(x>0.99) * (y>0.99)']), [3, 4])
        assert allc(nnz(gi['(x>0.99) * (y>0.99) * r']), [3])

    def test_global_eidx_conec0D(self, gi=None):
        nnz = np.nonzero
        allc = np.allclose
        if gi is None:
            gi = GlobalEIndex(self.conec0D, self.coors)
            gi.nidx.g.update(self.ngroups)
            self.test_global_nidx_conec0D(gi.nidx.g)
            self.test_nidx_conec0D(gi.nidx)
            gi.update(self.egroups0D)
            assert allc(nnz(gi.nidx.g['c5*opoint']), [9])
            assert allc(nnz(gi.nidx['c5*opoint']), [4])
        assert len(gi[:]) == 5
        assert allc(nnz(gi[:]), np.arange(5))
        assert allc(gi.pos, [0, 1, 2, 3, 4])
        with pytest.raises(KeyError):
            assert allc(nnz(gi['o']), [2, 3])
        assert allc(nnz(gi['c5']), [4])
        assert allc(nnz(gi['c1+c5']), [0, 4])
        assert allc(nnz(gi['c2+c3+c4']), [1, 2, 3])
        assert allc(nnz(gi['~(c2+c3+c4)^c5']), [0])
        assert allc(nnz(gi['c5*opoint']), [4])     # do we really need this?
        gi.nidx.g['conec2D'] = self.conec2D
        assert allc(nnz(gi['conec2D']), [0, 1, 2, 3])  # do we need this?
        assert allc(nnz(gi['conec2D*(c1+c2)']), [0, 1])  # do we need this?
        assert allc(nnz(gi['x>0.99']), [1, 3, 4])
        assert allc(nnz(gi['(x>0.99) * (y>0.99)']), [3])
        assert allc(nnz(gi['(x>0.99) * (y>0.99) * opoint']), [])

    def test_global_eidx_all(self):
        nnz = np.nonzero
        allc = np.allclose
        if not None:
            ni = GlobalNIndex(self.coors, self.conec2D)
            gi2D = GlobalEIndex(self.conec2D, ni)
            gi1D = GlobalEIndex(self.conec1D, ni)
            gi0D = GlobalEIndex(self.conec0D, ni)
            ni.update(self.ngroups)
            self.test_global_nidx_conec2D(ni)
            self.test_nidx_conec2D(gi2D.nidx)
            # self.test_nidx_conec1Da(gi1D.nidx)
            # self.test_nidx_conec0D(gi0D.nidx)
            gi2D.update(self.egroups2D)
            gi1D.update(self.egroups1D)
            gi0D.update(self.egroups0D)
        # test 2D
        assert allc(nnz(gi2D['q1*b + q1*l']), [])  # do we need this?
        # test 1D
        assert len(gi1D[:]) == 9
        assert allc(nnz(gi1D[:]), np.arange(8))
        assert allc(gi1D.pos, [0, 1, 2, 3, 4, 5, 6, 7, -1])
        assert allc(nnz(gi1D['q1*b + q1*l']), [0, 7])  # yes, we need this
        assert allc(nnz(gi1D['q1*o']), [])  # do we need this?
        assert allc(nnz(gi1D['c4+o']), [])  # do we need this?
        # test 0D
        assert len(gi0D[:]) == 5
        assert allc(nnz(gi0D[:]), np.arange(4))
        assert allc(gi0D.pos, [0, 1, 2, 3, -1])
        assert allc(nnz(gi0D['q1*c1']), [0])  # yes, we need this
        assert allc(nnz(gi0D['b*c1']), [0])  # yes, we need this
        assert allc(nnz(gi0D['b+opoint']), [0, 1])  # yes, we need this

    def test_eidx_create(self):
        allc = np.allclose
        nnz = np.nonzero
        # create from conec and coors
        with pytest.raises(ValueError):
            ei = EIndex(self.conec2D)
        ei = EIndex(self.conec2D, nidx=self.coors)
        ei.nidx.g.update(self.ngroups)
        self.test_global_nidx_conec2D(ei.nidx.g)
        self.test_nidx_conec2D(ei.nidx)
        ei.g.update(self.egroups2D)
        self.test_global_eidx_conec2D(ei.g)
        # create from conec and nindex
        ni = NIndex(self.coors, self.conec1D)
        ni.g.update(self.ngroups)
        ei = EIndex(self.conec2D, nidx=ni)
        self.test_global_nidx_conec1D(ei.nidx.g)
        assert allc(ei.nidx.lst, [0, 1, 3, 5, 7, 8])
        assert allc(ei.lst, [0, 7])
        ei.g.update(self.egroups2D)
        assert allc(nnz(ei['q4']), [1])
        assert allc(ei.subindex('q4').lst, [7])
        # create from conec and global nindex
        gni = GlobalNIndex(self.coors, self.conec1D)
        gni.update(self.ngroups)
        ei = EIndex(self.conec2D, nidx=gni)
        self.test_global_nidx_conec1D(ei.nidx.g)
        assert allc(ei.nidx.lst, [0, 1, 3, 5, 7, 8])
        assert allc(ei.lst, [0, 7])
        ei.g.update(self.egroups2D)
        assert allc(nnz(ei['q4']), [1])
        assert allc(ei.subindex('q4').lst, [7])
        # create from global eindex
        gei = GlobalEIndex(self.conec2D, gni)
        ei = EIndex(gei)
        self.test_global_nidx_conec1D(ei.nidx.g)
        assert allc(ei.nidx.lst, [0, 1, 3, 5, 7, 8])
        assert allc(ei.lst, [0, 7])
        ei.g.update(self.egroups2D)
        assert allc(nnz(ei['q4']), [1])
        assert allc(ei.subindex('q4').lst, [7])
        # create from global eindex and shadowing lst
        gei = GlobalEIndex(self.conec2D, gni)
        ei = EIndex(gei, lst=[0, 1, 2])
        self.test_global_nidx_conec1D(ei.nidx.g)
        assert allc(ei.nidx.lst, [0, 1, 3])
        assert allc(ei.lst, [0])
        ei.g.update(self.egroups2D)
        assert allc(nnz(ei['q1']), [0])
        assert allc(ei.subindex('q1').lst, [0])
        # create from global eindex and shadowing lst
        gei = GlobalEIndex(self.conec2D, gni)
        ei = EIndex(gei, lst=[0, 1, 2])
        self.test_global_nidx_conec1D(ei.nidx.g)
        assert allc(ei.nidx.lst, [0, 1, 3])
        assert allc(ei.lst, [0])
        ei.g.update(self.egroups2D)
        assert allc(nnz(ei['q1']), [0])
        assert allc(ei.subindex('q1').lst, [0])
        # create from local eindex
        with pytest.raises(ValueError):
            EIndex(ei, [0, 1, 2])

    def test_eidx_conec2D(self):
        nnz = np.nonzero
        allc = np.allclose
        ei = EIndex(self.conec2D, nidx=self.coors)
        ei.nidx.g.update(self.ngroups)
        self.test_global_nidx_conec2D(ei.nidx.g)
        self.test_nidx_conec2D(ei.nidx)
        ei.g.update(self.egroups2D)
        ei = ei.subindex([0, 1, 2, 4, 5, 6, 7])
        self.test_global_eidx_conec2D(ei.g)
        assert len(ei[:]) == 7
        assert allc(nnz(ei[:]), np.arange(7))
        assert allc(ei.pos, [0, 1, 2, -1, 3, 4, 5, 6])
        assert allc(nnz(ei['q2']), [2])
        # assert allc(nnz(ei.nidx['q2']), [1, 2, 4])   ##???
        assert allc(nnz(ei['q1+q2']), [0, 1, 2])
        assert allc(nnz(ei['q3+q4']), [3, 4, 5, 6])
        assert allc(nnz(ei['~(q1+q2)']), [3, 4, 5, 6])
        assert allc(nnz(ei['q1*cline']), [])     # do we really need this?
        ei.nidx.g['conec1D'] = self.conec1D
        assert allc(nnz(ei['conec1D']), [0, 6])  # do we really need this?
        assert allc(nnz(ei['conec1D*q4']), [6])  # do we really need this?
        assert allc(nnz(ei['x>0.99']), [2, 5, 6])
        assert allc(nnz(ei['(x>0.99) * (y>0.99)']), [5, 6])
        assert allc(nnz(ei['(x>0.99) * (y>0.99) * conec1D']), [6])

    def test_eidx_conec1D(self):
        nnz = np.nonzero
        allc = np.allclose
        ei = EIndex(self.conec1D, nidx=self.coors)
        ei.nidx.g.update(self.ngroups)
        self.test_global_nidx_conec1D(ei.nidx.g)
        self.test_nidx_conec1Da(ei.nidx)
        ei.update(self.egroups1D)
        ei = ei.subindex([0, 1, 2, 4, 5, 6, 7])
        self.test_global_eidx_conec1D(ei.g)
        assert allc(nnz(ei.nidx.g['r*cline']), [5])
        assert allc(nnz(ei.nidx['r*cline']), [4])
        assert len(ei[:]) == 7
        assert allc(nnz(ei[:]), np.arange(7))
        assert allc(ei.pos, [0, 1, 2, -1, 3, 4, 5, 6, -1])
        with pytest.raises(KeyError):
            assert allc(nnz(ei['q2']), [2, 3])
        assert allc(nnz(ei['l']), [5, 6])
        assert allc(nnz(ei['t+l']), [3, 4, 5, 6])
        assert allc(nnz(ei['b+r']), [0, 1, 2])
        assert allc(nnz(ei['~(b+r)']), [3, 4, 5, 6])
        assert allc(nnz(ei['l*cline']), [])     # do we really need this?
        ei.nidx.g['conec0D'] = self.conec0D
        assert allc(nnz(ei['conec0D']), [])  # do we really need this?
        assert allc(nnz(ei['conec0D*l']), [])  # do we really need this?
        assert allc(nnz(ei['x>0.99']), [1, 2, 3])
        assert allc(nnz(ei['(x>0.99) * (y>0.99)']), [3])
        assert allc(nnz(ei['(x>0.99) * (y>0.99) * t']), [3])

    def test_eidx_conec0D(self):
        nnz = np.nonzero
        allc = np.allclose
        ei = EIndex(self.conec0D, nidx=self.coors)
        ei.nidx.g.update(self.ngroups)
        self.test_global_nidx_conec0D(ei.nidx.g)
        self.test_nidx_conec0D(ei.nidx)
        ei.g.update(self.egroups0D)
        ei = ei.subindex([0, 1, 2, 4])
        self.test_global_eidx_conec0D(ei.g)
        assert allc(nnz(ei.nidx.g['c5*opoint']), [9])
        assert allc(nnz(ei.nidx['c5*opoint']), [3])
        assert len(ei[:]) == 4
        assert allc(nnz(ei[:]), np.arange(4))
        assert allc(ei.pos, [0, 1, 2, -1, 3])
        with pytest.raises(KeyError):
            assert allc(nnz(ei['o']), [2, 3])
        assert allc(nnz(ei['c5']), [3])
        assert allc(nnz(ei['c1+c5']), [0, 3])
        assert allc(nnz(ei['c2+c3+c4']), [1, 2])
        assert allc(nnz(ei['~(c2+c3+c4)^c5']), [0])
        assert allc(nnz(ei['c5*opoint']), [3])     # do we really need this?
        ei.nidx.g['conec2D'] = self.conec2D
        assert allc(nnz(ei['conec2D']), [0, 1, 2])  # do we need this?
        assert allc(nnz(ei['conec2D*(c1+c2)']), [0, 1])  # do we need this?
        assert allc(nnz(ei['x>0.99']), [1, 3])
        assert allc(nnz(ei['(x>0.99) * (y>0.99)']), [])
        assert allc(nnz(ei['(x>0.99) * (y>0.99) * opoint']), [])

    def test_eidx_all(self):
        nnz = np.nonzero
        allc = np.allclose
        if not None:
            gni = GlobalNIndex(self.coors, self.conec2D)
            gni.update(self.ngroups)
            gei2D = GlobalEIndex(self.conec2D, gni)
            gei1D = GlobalEIndex(self.conec1D, gni)
            gei0D = GlobalEIndex(self.conec0D, gni)
            gei2D.update(self.egroups2D)
            gei1D.update(self.egroups1D)
            gei0D.update(self.egroups0D)
            ei2D = EIndex(gei2D)
            ei1D = EIndex(gei1D)
            ei0D = EIndex(gei0D)
        # test 2D
        assert allc(nnz(ei2D['q1*b + q1*l']), [])  # do we need this?
        ei = ei2D.subindex([1, 2, 3])
        assert allc(nnz(ei['q1*b + q1*l']), [])  # do we need this?
        ei = ei2D.subindex('q1+q2')
        assert allc(nnz(ei['q1+q3']), [0, 1])  # do we need this?
        # test 1D
        assert len(ei1D[:]) == 8
        assert allc(nnz(ei1D[:]), np.arange(8))
        assert allc(ei1D.pos, [0, 1, 2, 3, 4, 5, 6, 7, -1])
        assert allc(nnz(ei1D['q1*b + q1*l']), [0, 7])  # yes, we need this
        assert allc(nnz(ei1D['q1*o']), [])  # do we need this?
        assert allc(nnz(ei1D['c4+o']), [])  # do we need this?
        ei = ei1D.subindex('b')
        assert allc(ei.pos, [0, 1, -1, -1, -1, -1, -1, -1, -1])
        assert allc(nnz(ei['q1*b + q1*l']), [0])  # yes, we need this
        # test 0D
        assert len(ei0D[:]) == 4
        assert allc(nnz(ei0D[:]), np.arange(4))
        assert allc(ei0D.pos, [0, 1, 2, 3, -1])
        assert allc(nnz(ei0D['q1*c1']), [0])  # yes, we need this
        assert allc(nnz(ei0D['b*c1']), [0])  # yes, we need this
        assert allc(nnz(ei0D['b+opoint']), [0, 1])  # yes, we need this

    def test_empty_key(self):
        nnz = np.nonzero
        allc = np.allclose
        ei = EIndex(self.conec2D, nidx=self.coors)
        assert allc(nnz(~ei[[]]), np.arange(ei.n))
        assert allc(nnz(~ei.nidx[[]]), np.arange(ei.nidx.n))
        assert allc(nnz(~ei.g[[]]), np.arange(ei.g.n))
        assert allc(nnz(~ei.nidx.g[[]]), np.arange(ei.nidx.g.n))

    def test_global_nidx_assign(self):
        i = GlobalNIndex(self.coors, self.conec2D)
        i['aaa'] = [0, 1]
        i['bbb'] = np.ones(i.n, dtype=np.bool)


if __name__ == '__main__':
    # import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file
